@@ -2,6 +2,11 @@
 Description: Processes model results for visualization
 
 Uses methods:
+    - :func:`save_result`:              Saves a given result variable (endclasses or mdlhists) to a file filename. 
+    - :func:`load_result`:              Loads a given (endclasses or mdlhists) results dictionary from a (pickle/csv/json) file.
+    - :func:`load_results`:             Loads endclass/mdlhist results from a given folder 
+    - :func:`flatten_hist`:             Recursively creates a flattenned (single-level dictionary) history of the given nested model history
+    - :func:`renest_flattened_hist:     Re-nests a flattened history.  
     - :func:`hists`:                    Processes a model histories for each scenario into results histories by comparing the states over time in each scenario with the states in the nominal scenario.
     - :func:`hist`:                     Compares model history with the nominal model history over time to make a history of degradation.
         - :func:`fxnhist`:              Compares the history of function states in mdlhist over time.
@@ -35,9 +40,9 @@ Also used for graph heatmaps, which use the results history to map results histo
 #Created: November 2019 (Refactored April 2020)
 
 import copy
-import networkx as nx
 import numpy as np
 import pandas as pd
+import sys,os
 from ordered_set import OrderedSet
 from fmdtools.faultsim.propagate import cut_mdlhist
 from scipy.stats import bootstrap
@@ -503,12 +508,14 @@ def nan_to_x(metric, x=0.0):
 def expected(endclasses, metric):
     """Calculates the expected value of a given metric in endclasses using the rate variable in endclasses"""
     return sum([e[metric]*e['rate'] for k,e in endclasses.items() if not np.isnan(e[metric])])
-def average(endclasses, metric):
+def average(endclasses, metric, empty_as='nan'):
     """Calculates the average value of a given metric in endclasses"""
-    return np.mean([e[metric] for k,e in endclasses.items() if not np.isnan(e[metric])])
+    ecs = [e[metric] for k,e in endclasses.items() if not np.isnan(e[metric])]
+    if len(ecs)>0 or empty_as=='nan':   return np.mean(ecs)
+    else:                               return empty_as
 def percent(endclasses, metric):
     """Calculates the percentage of a given indicator variable being True in endclasses"""
-    return sum([int(bool(e[metric])) for k,e in endclasses.items() if not np.isnan(e[metric])])/len(endclasses)
+    return sum([int(bool(e[metric])) for k,e in endclasses.items() if not np.isnan(e[metric])])/(len(endclasses)+1e-16)
 def rate(endclasses, metric):
      """Calculates the rate of a given indicator variable being True in endclasses using the rate variable in endclasses"""
      return sum([int(bool(e[metric]))*e['rate'] for k,e in endclasses.items() if not np.isnan(e[metric])])
@@ -587,3 +594,300 @@ def bootstrap_confidence_interval(data, method=np.mean, return_anyway=False, **k
     elif return_anyway: return method(data), method(data), method(data)
     else: raise Exception("All data are the same!")
 
+def file_check(filename, overwrite):
+    """Check if files exists and whether to overwrite the file"""
+    if os.path.exists(filename):
+        if not overwrite: raise Exception("File already exists: "+filename)
+        else:                   
+            print("File already exists: "+filename+", writing anyway...")
+            os.remove(filename)
+
+def save_result(variable, filename, filetype="", overwrite=False, result_id=''):
+    """
+    Saves a given result variable (endclasses or mdlhists) to a file filename. 
+    Files can be saved as pkl, csv, or json.
+
+    Parameters
+    ----------
+    variable : dict
+        Results dictionary (endclasses or mdlhists)
+    filename : str
+        File name for the file. Can be nested in a folder if desired.
+    filetype : str, optional
+        Optional specifier of file type (if not included in filename). The default is "".
+    overwrite : bool, optional
+        Whether to overwrite existing files with this name. The default is False.
+    result_id : str, optional
+        For individual results saving. Places an identifier for the result in the file. The default is ''.
+    """
+    import dill, json, csv
+    file_check(filename, overwrite)
+    
+    if "/" in filename:
+        last_split_index = filename.rfind("/")
+        foldername = filename[:last_split_index]
+        if not os.path.exists(foldername): os.makedirs(foldername)
+    
+    filetype = auto_filetype(filename, filetype)
+    if filetype=='pickle':
+        with open(filename, 'wb') as file_handle:
+            if result_id: variable = {result_id:variable}
+            dill.dump(variable, file_handle)
+    elif filename[-4:]=='.csv': # add support for nested dict mdlhist using flatten_hist?
+        variable = flatten_hist(variable)
+        with open(filename, 'w', newline='') as file_handle:
+            writer = csv.writer(file_handle)
+            if result_id: writer.writerow([result_id])
+            writer.writerow(variable.keys())
+            if isinstance([*variable.values()][0], np.ndarray):
+                writer.writerows(zip(*variable.values()))
+            else:
+                writer.writerow([*variable.values()])
+    elif filename[-5:]=='.json':
+        with open(filename, 'w', encoding='utf8') as file_handle:
+            variable = flatten_hist(variable)
+            new_variable = {}
+            for key in variable:
+                if isinstance(variable[key], np.ndarray):
+                    new_variable[str(key)] =  [var.item() for var in variable[key]]
+                else:
+                    new_variable[str(key)] =  variable[key]
+            if result_id: new_variable = {result_id:new_variable}
+            strs = json.dumps(new_variable, indent=4, sort_keys=True, separators=(',', ': '), ensure_ascii=False)
+            file_handle.write(str(strs))
+    else:
+        raise Exception("Invalid File Type")
+    file_handle.close()
+        
+def load_result(filename, filetype="", renest_dict=True, indiv=False):
+    """
+    Loads a given (endclasses or mdlhists) results dictionary from a (pickle/csv/json) file.
+    e.g. a file saved using process.save_result or save_args in propagate functions.
+
+    Parameters
+    ----------
+    filename : str
+        Name of the file.
+    filetype : str, optional
+        Use to specify a filetype for the file (if not included in the filename). The default is "".
+    renest_dict : bool, optional
+        Whether to return . The default is True.
+    indiv : bool, optional
+        Whether the result is an individual file (e.g. in a folder of results for a given simulation). 
+        The default is False.
+
+    Returns
+    -------
+    resultdict : dict
+        Corresponding results dictionary from the file.
+    """
+    import dill, json, csv, pandas
+    if not os.path.exists(filename): raise Exception("File does not exist: "+filename)
+    filetype = auto_filetype(filename, filetype)
+    if filetype=='pickle':
+        with open(filename, 'rb') as file_handle:
+            resultdict = dill.load(file_handle)
+        file_handle.close()
+    elif filetype=='csv': # add support for nested dict mdlhist using flatten_hist?
+        if indiv:   resulttab = pandas.read_csv(filename, skiprows=1)
+        else:           resulttab = pandas.read_csv(filename)
+        resultdict = resulttab.to_dict("list")
+        resultdict = clean_resultdict_keys(resultdict)
+        for key in resultdict:
+            if len(resultdict[key])==1 and isinstance(resultdict[key], list):
+                resultdict[key] = resultdict[key][0]
+            else: resultdict[key] = np.array(resultdict[key])             
+        if renest_dict: resultdict = nest_flattened_hist(resultdict)
+        if indiv: 
+            scenname = [*pandas.read_csv(filename, nrows=0).columns][0]
+            resultdict = {scenname: resultdict}
+    elif filetype=='json':
+        with open(filename, 'r', encoding='utf8') as file_handle:
+            loadeddict = json.load(file_handle)
+            if indiv:   
+                key = [*loadeddict.keys()][0]
+                loadeddict = loadeddict[key]
+                loadeddict= {key+", "+innerkey:values for innerkey, values in loadeddict.items()}
+                resultdict = clean_resultdict_keys(loadeddict)
+            else:       resultdict = clean_resultdict_keys(loadeddict)
+            
+            if renest_dict: resultdict = nest_flattened_hist(resultdict)
+        file_handle.close()
+    else:
+        raise Exception("Invalid File Type")
+    return resultdict
+def clean_resultdict_keys(resultdict_dirty):
+    """
+    Helper function for recreating results dictionary keys (tuples) from a dictionary loaded from a file (where keys are strings)
+    (used in csv/json results)
+
+    Parameters
+    ----------
+    resultdict_dirty : dict
+        Results dictionary where keys are strings
+
+    Returns
+    -------
+    resultdict : dict
+        Results dictionary where keys are tuples
+    """
+    resultdict = {}
+    for key in resultdict_dirty:
+        newkey = tuple(key.replace("'","").replace("(","").replace(")","").split(", "))
+        if any(['t=' in strs for strs in  newkey]):
+            joinfirst = [ind for ind, strs in enumerate(newkey) if 't=' in strs][0] +1
+        else:                           joinfirst=0
+        if joinfirst==2:
+            newkey = tuple([", ".join(newkey[:2])])+newkey[2:]
+        elif joinfirst>2:
+            nomscen = newkey[:joinfirst-2]
+            faultscen = tuple([", ".join(newkey[joinfirst-2:joinfirst])])
+            vals = newkey[joinfirst:]
+            newkey = nomscen+faultscen+vals
+        resultdict[newkey]=resultdict_dirty[key]
+    return resultdict
+
+def load_results(folder, filetype, renest_dict=True):
+    """
+    Loads endclass/mdlhist results from a given folder 
+    (e.g., that have been saved from multi-scenario propagate methods with 'indiv':True)
+
+    Parameters
+    ----------
+    folder : str
+        Name of the folder. Must be in the current directory
+    filetype : str
+        Type of files in the folder ('pickle', 'csv', or 'json')
+    renest_dict : bool, optional
+        Whether to return result as a nested dict (as opposed to a flattenned dict). 
+        The default is True.
+
+    Returns
+    -------
+    resultdict : dict
+        endclasses/mdlhists result dictionary reconstructed by the files in the folder.
+    """
+    files = os.listdir(folder)
+    files_toread = []
+    for file in files:
+        read_filetype = auto_filetype(file)
+        if read_filetype==filetype:
+            files_toread.append(file)
+    resultdict = {}
+    for filename in files_toread:
+        resultdict.update(load_result(folder+'/'+filename, filetype, renest_dict=renest_dict, indiv=True))
+    return resultdict
+
+def auto_filetype(filename, filetype=""):
+    """Helper function that automatically determines the filetype (pickle, csv, or json) of a given filename"""
+    if not filetype:
+        if '.' not in filename: raise Exception("No file extension")
+        if filename[-4:]=='.pkl':       filetype="pickle"
+        elif filename[-4:]=='.csv':     filetype="csv"     
+        elif filename[-5:]=='.json':    filetype="json"
+        else: raise Exception("Invalid File Type in: "+filename+", ensure extension is pkl, csv, or json ")
+    return filetype
+def create_indiv_filename(filename, indiv_id, splitchar='_'):
+    """Helper file that creates an individualized name for a file given the general filename and an individual id"""
+    filename_parts = filename.split(".")
+    filename_parts.insert(1,'.')
+    filename_parts.insert(1,splitchar+indiv_id)   
+    return "".join(filename_parts)
+
+def get_hist_memory(hist):
+    """
+    Determines the memory usage of a given history and profiles by 
+
+    Parameters
+    ----------
+    hist : TYPE
+        DESCRIPTION.
+
+    Returns
+    -------
+    mem_total : int
+        Total memory usage of the history (in bytes)
+    mem_profile : dict
+        Memory usage of each construct of the model history (in bytes)
+    """
+    fhist = flatten_hist(hist)
+    mem_total = 0
+    mem_profile = dict.fromkeys(fhist.keys())
+    for k,h in fhist.items():
+        if np.issubdtype(h.dtype, np.number) or np.issubdtype(h.dtype, np.flexible) or np.issubdtype(h.dtype, np.bool_):
+            mem=h.nbytes
+        else:
+            mem=0
+            for entry in h:
+                mem+=sys.getsizeof(entry)
+        mem_total+=mem
+        mem_profile[k]=mem
+    return mem_total, mem_profile
+
+def get_flat_hist_slice(flathist,t_ind=0):
+    """
+    Returns a dictionary of values from a given (flat) mdlhist at t_ind
+    """
+    slice_dict = dict.fromkeys(flathist)
+    for key, arr in flathist.items():
+        slice_dict[key]=flathist[key][t_ind]
+    return slice_dict
+
+def flatten_hist(hist, newhist = False, prevname=(), to_include='all'):
+    """
+    Recursively creates a flattenned history of the given nested model history
+
+    Parameters
+    ----------
+    hist : dict
+        Model history (e.g., from faultsim.propagate.nominal).
+    newhist : dict, optional
+        Flattened Model History (used when called recursively). The default is False.
+    prevname : tuple, optional
+        Current key of the flattened history (used when called recursively). The default is ().
+    to_include : str/list/dict, optional
+        What attributes to include in the dict. The default is 'all'. Can be of form
+        - list e.g. ['att1', 'att2', 'att3'] to include the given attributes
+        - dict e.g. fxnflowvals {'flow1':['att1', 'att2'], 'fxn1':'all', 'fxn2':['comp1':all, 'comp2':['att1']]}
+        - str e.g. 'att1' for attribute 1 or 'all' for all attributes
+    Returns
+    -------
+    newhist : dict
+        Flattened model history of form: {(fxnflow, ..., attname):array(att)}
+    """
+    if newhist==False: newhist = dict()
+    for att, val in hist.items():
+        newname = prevname+tuple([att])
+        if type(val)==dict: 
+            if type(to_include)==list and att in to_include: new_to_include = 'all'
+            elif type(to_include)==set and att in to_include: new_to_include = 'all'
+            elif type(to_include)==dict and att in to_include: new_to_include = to_include[att]
+            elif type(to_include)==str and att== to_include: new_to_include = 'all'
+            elif to_include =='all': new_to_include='all'
+            elif att in ['functions', 'flows']: new_to_include = to_include
+            else: new_to_include= False
+            if new_to_include: flatten_hist(val, newhist, newname, new_to_include)
+        elif to_include=='all' or att in to_include: 
+            if len(newname)==1: newhist[newname[0]] = val
+            else:               newhist[newname] = val
+    return newhist
+
+def nest_flattened_hist(hists, prefix = ()):
+    """
+    Re-nests a flattened history.   
+
+    Parameters
+    ----------
+    hists : dict
+        Flattened Model history (e.g. from flatten_hist)
+    """
+    newhist = {}
+    key_options = set([h[0] for h in hists.keys()])
+    for key in key_options:
+        if (key,) in hists:     newhist[key] = hists[(key,)]
+        else:
+            subdict = {histkey[1:]:val for histkey, val in hists.items() if key==histkey[0]}                       
+            newhist[key] = nest_flattened_hist(subdict)
+    return newhist
+            
+
